# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from tarina.lang.model import LangItem, LangModel


class EntariPluginInspectScene:
    name: LangItem = LangItem("entari::plugin_inspect", "scene.name")
    direct: LangItem = LangItem("entari::plugin_inspect", "scene.direct")
    text: LangItem = LangItem("entari::plugin_inspect", "scene.text")
    voice: LangItem = LangItem("entari::plugin_inspect", "scene.voice")
    category: LangItem = LangItem("entari::plugin_inspect", "scene.category")


class EntariPluginInspect:
    description: LangItem = LangItem("entari::plugin_inspect", "description")
    platform: LangItem = LangItem("entari::plugin_inspect", "platform")
    self: LangItem = LangItem("entari::plugin_inspect", "self")
    scene = EntariPluginInspectScene
    user: LangItem = LangItem("entari::plugin_inspect", "user")
    channel: LangItem = LangItem("entari::plugin_inspect", "channel")
    guild: LangItem = LangItem("entari::plugin_inspect", "guild")
    invalid: LangItem = LangItem("entari::plugin_inspect", "invalid")


class Lang(LangModel):
    entari_plugin_inspect = EntariPluginInspect
