# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from pathlib import Path

from tarina.lang import lang

lang.load(Path(__file__).parent)

from .model import Lang as Lang
